<?php

namespace Drupal\chaching\Controller;

use DateTime;
use DateTimeZone;
use Drupal\Component\Utility\Html;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheableJsonResponse;
use Drupal\Core\Cache\CacheableResponse;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Link;
use Drupal\Core\Serialization\Yaml;
use Drupal\Core\Template\Attribute;
use Drupal\Core\Url;
use Exception;
use stdClass;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Controller for Cha-ching pages.
 */
class ChachingController extends ControllerBase {

  /**
   * Database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * Language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;

  /**
   * Request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static($container->get('database'), $container->get('language_manager'), $container->get('request_stack'));
  }

  /**
   * {@inheritdoc}
   */
  final public function __construct($database, $language_manager, $request_stack) {
    $this->database = $database;
    $this->languageManager = $language_manager;
    $this->requestStack = $request_stack;
  }

  /**
   * IPN callback.
   */
  public function ipn() {
    $ipn = $this->requestStack->getCurrentRequest()->request->all();
    $validated = $this->validate($ipn);
    if (!$validated) {
      throw new Exception('IPN incoming NOT VERIFIED');
    }
    $receiver_emails = $this->config('chaching.settings')->get('receiver_email');
    if (!$receiver_emails) {
      throw new Exception('PayPal receiver email not configured.');
    }
    if (array_search($ipn['receiver_email'], $receiver_emails) === FALSE) {
      throw new Exception('PayPal IPN contains invalid receiver email address.');
    }
    $schema = drupal_get_module_schema('chaching', 'chaching_paypal_ipns');
    foreach ($ipn as $key => $value) {
      if (!isset($schema['fields'][$key])) {
        $this->getLogger('chaching')->error('IPN unknown field ignored: %key => %value', ['%key' => $key, '%value' => $value]);
        unset($ipn[$key]);
      }
    }
    $ipn['timestamp'] = time();
    $ipn['ipn'] = Yaml::encode($this->requestStack->getCurrentRequest()->request->all());
    if (!empty($ipn['payment_date'])) {
      $ipn['payment_timestamp'] = strtotime($ipn['payment_date']);
    }
    $ipn['verification_result'] = $validated;
    unset($ipn['processed']);
    $this->database->insert('chaching_paypal_ipns')
      ->fields($ipn)
      ->execute();
    Cache::invalidateTags(['chaching']);
    return [
      '#markup' => 'IPN: Only PayPal will ever see this page - humans go away!',
    ];
  }

  /**
   * Check with PayPal that this is a real IPN.
   *
   * Without this check, IPNs may be spoofed.
   *
   * @param array $post
   *   The key-value pairs from an incoming IPN (presumably from $_POST).
   *
   * @return bool
   *   TRUE if PayPal successfully validated the request.  This will not happen
   *   if there was an error communicating with the service, the IPN did not
   *   originate from PayPal, or it was tampered with in any way.
   */
  public function validate(array $post) {
    // Build the IPN validation request from the $post fields.
    $post['cmd'] = '_notify-validate';
    $req = http_build_query($post);
    $hostname = empty($post['test_ipn']) ? 'www.paypal.com' : 'www.sandbox.paypal.com';
    $ch = curl_init("https://$hostname/cgi-bin/webscr");
    curl_setopt($ch, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
    curl_setopt($ch, CURLOPT_POST, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $req);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 1);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 2);
    curl_setopt($ch, CURLOPT_FORBID_REUSE, 1);
    curl_setopt($ch, CURLOPT_HTTPHEADER, ['Connection: Close']);
    $res = curl_exec($ch);
    curl_close($ch);
    return trim($res) === 'VERIFIED';
  }

  /**
   * Donations API callback.
   */
  public function donations($type = 'total', $period = 'mtd', $format = 'json', $filter = '') {
    $timezone = new DateTimeZone('America/Los_Angeles');
    $date = new DateTime('now', $timezone);
    $output = new stdClass();
    $output->end = $date->format('r');
    switch (strtolower($period)) {
      case 'mtd':
        $date->setDate($date->format('Y'), $date->format('n'), 1)->setTime(0, 0);
        break;

      case '1m':
        $date->modify('-1 month');
        break;

      case 'ytd':
        $date->setDate($date->format('Y'), 1, 1)->setTime(0, 0);
        break;

      case '1y':
        $date->modify('-1 year');
        break;

      case 'all':
        $date->setDate(2006, 12, 17)->setTime(12, 0);
        break;

      default:
        throw new NotFoundHttpException();
    }
    $output->start = $date->format('r');
    switch (strtolower($filter)) {
      case 'noisetor':
        $filter_query = "AND item_name = 'Noisebridge Tor Node Project'";
        break;

      default:
        $filter_query = "AND item_name <> 'Noisebridge Tor Node Project'";
    }
    switch (strtolower($type)) {
      case 'total':
        $total = $this->database->query("SELECT SUM(mc_gross) AS total_gross, SUM(mc_gross - mc_fee) AS total_net, COUNT(*) AS count FROM {chaching_paypal_ipns} WHERE payment_status = :status AND timestamp >= :timestamp $filter_query", [':status' => 'Completed', ':timestamp' => $date->getTimestamp()])->fetchObject();
        $output->total_gross = (float) $total->total_gross;
        $output->total_net = (float) $total->total_net;
        $output->count = (int) $total->count;
        break;

      case 'list':
        $output->donations = [];
        $payment_date = new DateTime('now', $timezone);
        $result = $this->database->query("SELECT timestamp, mc_gross, mc_fee, mc_gross - mc_fee AS payment_net, id, txn_type FROM {chaching_paypal_ipns} WHERE payment_status = :status AND timestamp >= :timestamp $filter_query", [':status' => 'Completed', ':timestamp' => $date->getTimestamp()]);
        foreach ($result as $donation) {
          $donation->timestamp = (int) $donation->timestamp;
          $donation->id = (int) $donation->id;
          $payment_date->setTimestamp($donation->timestamp);
          $donation->payment_date = $payment_date->format('r');
          $donation->mc_fee = (float) $donation->mc_fee;
          $donation->mc_gross = (float) $donation->mc_gross;
          $donation->payment_net = (float) $donation->payment_net;
          $donation->recurring = ($donation->txn_type == 'subscr_payment');
          $output->donations[] = $donation;
        }
        $output->count = count($output->donations);
        break;

      default:
        throw new NotFoundHttpException();
    }
    switch (strtolower($format)) {
      case 'json':
        $response = new CacheableJsonResponse($output, 200, ['Access-Control-Allow-Origin' => '*']);
        $response->addCacheableDependency(CacheableMetadata::createFromRenderArray(['#cache' => ['tags' => ['chaching']]]));
        return $response;

      case 'rss':
        $output->filter = $filter;
        return $this->rss($output);

      default:
        throw new NotFoundHttpException();
    }
  }

  /**
   * Graph page callback.
   */
  public function graph($period = '1m', $filter = '') {
    return [
      '#type' => 'html_tag',
      '#tag' => 'div',
      '#attributes' => [
        'id' => 'chaching-plot',
      ],
      '#attached' => [
        'drupalSettings' => [
          'chaching' => ['period' => $period, 'filter' => $filter],
        ],
        'feed' => [
          [Url::fromUri("internal:/v1/donations/list/$period/rss/$filter")->toString(), 'Noisebridge donations RSS'],
        ],
        'library' => [
          'chaching/graph',
        ],
      ],
    ];
  }

  /**
   * Docs page callback.
   */
  public function docs() {
    $parameters = [
      ['type' => 'total', 'period' => '1m', 'format' => 'json'],
      ['type' => 'total', 'period' => '1y', 'format' => 'json'],
      ['type' => 'list', 'period' => '1m', 'format' => 'json'],
      ['type' => 'list', 'period' => '1y', 'format' => 'json'],
      [
        'type' => 'total',
        'period' => '1m',
        'format' => 'json',
        'filter' => 'noisetor',
      ],
      [
        'type' => 'total',
        'period' => '1y',
        'format' => 'json',
        'filter' => 'noisetor',
      ],
      [
        'type' => 'list',
        'period' => '1m',
        'format' => 'json',
        'filter' => 'noisetor',
      ],
      [
        'type' => 'list',
        'period' => '1y',
        'format' => 'json',
        'filter' => 'noisetor',
      ],
    ];
    $descriptions = [
      'Total donations to Noisebridge general fund in past month.',
      'Total donations to Noisebridge general fund in past year.',
      'List of donations to Noisebridge general fund in past month.',
      'List of donations to Noisebridge general fund in past year.',
      'Total Noisetor donations in past month.',
      'Total Noisetor donations in past year.',
      'List of Noisetor donations in past month.',
      'List of Noisetor donations in past year.',
    ];
    foreach ($parameters as $key => $item) {
      $url = Url::fromRoute('chaching.donations', $item, ['absolute' => TRUE]);
      $json[] = [
        Link::fromTextAndUrl($url->toString(), $url)->toRenderable(),
        ['#markup' => '<br />' . $descriptions[$key]],
      ];
    }
    $output[] = [
      '#theme' => 'item_list',
      '#items' => $json,
      '#title' => 'JSON format',
    ];
    $parameters = [
      ['type' => 'list', 'period' => '1m', 'format' => 'rss'],
      [
        'type' => 'list',
        'period' => '1m',
        'format' => 'rss',
        'filter' => 'noisetor',
      ],
    ];
    $descriptions = [
      'List of donations to Noisebridge general fund in past month.',
      'List of Noisetor donations in past month.',
    ];
    foreach ($parameters as $key => $item) {
      $url = Url::fromRoute('chaching.donations', $item, ['absolute' => TRUE]);
      $rss[] = [
        Link::fromTextAndUrl($url->toString(), $url)->toRenderable(),
        ['#markup' => '<br />' . $descriptions[$key]],
      ];
    }
    $output[] = [
      '#theme' => 'item_list',
      '#items' => $rss,
      '#title' => 'RSS format',
    ];
    $output[] = [
      '#markup' => '<blockquote><pre>' . file_get_contents(drupal_get_path('module', 'chaching') . '/sample-code.js') . '</pre></blockquote>',
    ];
    return $output;
  }

  /**
   * Outputs data in RSS format.
   */
  public function rss($var = NULL) {
    global $base_url;
    $language = $this->languageManager->getCurrentLanguage();
    $items = '';
    if (isset($var->donations)) {
      foreach (array_reverse($var->donations) as $item) {
        $description = '$' . $item->mc_gross . ' donation received';
        $extra = [
          'pubDate' => gmdate('r', $item->timestamp),
          'guid' => $base_url . '/ipn/' . $item->id,
        ];
        $items .= $this->formatRssItem('Cha-ching! ' . $description, $base_url, $description . '.', $extra);
      }
    }
    switch ($var->filter) {
      case 'noisetor':
        $title = 'Noisetor donations';
        $description = 'Latest donations to the Noisetor project.';
        break;

      default:
        $title = 'Noisebridge donations';
        $description = 'Latest donations to Noisebridge.';
    }
    $output = '<?xml version="1.0" encoding="utf-8"?>
<rss version="2.0" xml:base="' . $base_url . '" xmlns:atom="http://www.w3.org/2005/Atom">
';
    $output .= $this->formatRssChannel($title, $base_url, $description, $items, $language->getId(), [
      [
        'key' => 'atom:link',
        'attributes' => [
          'href' => $base_url . $this->requestStack->getCurrentRequest()->getRequestUri(),
          'rel' => 'self',
          'type' => 'application/rss+xml',
        ],
      ],
    ]) . '</rss>';
    $response = new CacheableResponse($output, 200, ['Access-Control-Allow-Origin' => '*', 'Content-Type' => 'application/rss+xml; charset=utf-8']);
    $response->addCacheableDependency(CacheableMetadata::createFromRenderArray(['#cache' => ['tags' => ['chaching']]]));
    return $response;
  }

  /**
   * Formats RSS channel.
   */
  public function formatRssChannel($title, $link, $description, $items, $langcode = NULL, $args = []) {
    $language_content = $this->languageManager->getCurrentLanguage();
    $langcode = $langcode ? $langcode : $language_content->getId();
    $output = "<channel>\n";
    $output .= ' <title>' . Html::escape($title) . "</title>\n";
    $output .= ' <link>' . Html::escape($link) . "</link>\n";

    // The RSS 2.0 "spec" doesn't indicate HTML can be used in the description.
    // We strip all HTML tags, but need to prevent double encoding from properly
    // escaped source data (such as &amp becoming &amp;amp;).
    $output .= ' <description>' . Html::escape(Html::decodeEntities(strip_tags($description))) . "</description>\n";
    $output .= ' <language>' . Html::escape($langcode) . "</language>\n";
    $output .= $this->formatXmlElements($args);
    $output .= $items;
    $output .= "</channel>\n";
    return $output;
  }

  /**
   * Formats RSS item.
   */
  public function formatRssItem(string $title, string $link, string $description, array $args = []) {
    $output = "<item>\n";
    $output .= ' <title>' . Html::escape($title) . "</title>\n";
    $output .= ' <link>' . Html::escape($link) . "</link>\n";
    $output .= ' <description>' . Html::escape($description) . "</description>\n";
    $output .= $this->formatXmlElements($args);
    $output .= "</item>\n";
    return $output;
  }

  /**
   * Formats XML elements.
   */
  public function formatXmlElements(array $array) {
    $output = '';
    foreach ($array as $key => $value) {
      if (is_numeric($key)) {
        if ($value['key']) {
          $output .= ' <' . $value['key'];
          if (isset($value['attributes']) && is_array($value['attributes'])) {
            $output .= new Attribute($value['attributes']);
          }
          if (isset($value['value']) && $value['value'] != '') {
            $output .= '>' . (is_array($value['value']) ? $this->formatXmlElements($value['value']) : (!empty($value['encoded']) ? $value['value'] : Html::escape($value['value']))) . '</' . $value['key'] . ">\n";
          }
          else {
            $output .= " />\n";
          }
        }
      }
      else {
        $output .= ' <' . $key . '>' . (is_array($value) ? $this->formatXmlElements($value) : Html::escape($value)) . "</{$key}>\n";
      }
    }
    return $output;
  }

}
